# -*- coding: utf-8 -*-
"""
A sober python base library
"""
# SPDX-License-Identifier: LGPL-3.0

__version__ = "0.8.6"

from .util import *   # noqa: F403, F401
from .obj  import *   # noqa: F403, F401



# -*- coding: utf-8 -*-
"""

.. |br| raw:: html

   <br />

SYNOPSIS
========

::

    from amethyst.core import Object, Attr

    class MyObject(Object):
        # Attr() defines properties and will include in serialization.
        foo = Attr(int)
        bar = Attr(str).strip()

        # foo and bar will be automatically extracted from kwargs.
        # .other will not be serialized by .toJSON()
        def __init__(self, other=None, **kwargs):
            super().__init__(**kwargs)
            self.other = other


    # ...
    myobj = MyObject(foo=23, other="Hello")
    myobj.toJSON()   # { "__my.module.MyObject__": { "foo": 23 } }

    myobj = MyObject()
    myobj.fromJSON('{ "foo": 23, "bar": " plugh  " }')
    print(myobj.bar)      # "plugh"  (no spaces)



DESCRIPTION
===========

:py:class:`Object` implements the dictionary interface and stores
everything in self.dict.

Subclasses can define :py:class:`Attr` which will have properties defined
as shortcuts to read and write keys in the dictionary.

By storing all attributes in a dict, we can be trivially serialized.
toJSON() and fromJSON() methods exist to help with this, and should be used
for all JSON serialization since they will correctly handle `set()` and
other values (see the :py:func:`Object.JSONEncoder` and
:py:func:`Object.JSONObjectHook` methods). Additionally, the JSON methods
will perform automatic validation based on type information passed to the
:py:class:`Attr` objects and will ensure that it is loading data for the
correct class and that no unexpected keys are present.

"""
# SPDX-License-Identifier: LGPL-3.0
from __future__ import division, absolute_import, print_function, unicode_literals

__all__ = """
Object
Attr
AmethystException
  DuplicateAttributeException
  ImmutableObjectException
register_amethyst_type
amethyst_deflate
amethyst_inflate
""".split()

import inspect
import json
import numbers
import six
import warnings

from .util import coalesce, smartmatch, get_class


class AmethystException(Exception): pass
class ImmutableObjectException(AmethystException): pass
class DuplicateAttributeException(AmethystException): pass


UNIQUE1 = object()
UNIQUE2 = object()


class Attr(object):
    """
    Base class for Amethyst Object Attributes

    Attribute descriptions primarily consist of a function which takes in a
    value and either returns a (possibly modified) value or else raises a
    :py:exc:`ValueError`. Python's standard object constructors generally work well,
    though beware that `str` will generally accept anything. ::

        foo = Attr(int)                   # Coerce to int (strict parsing)
        foo = Attr(float).int()           # Parse via float, but then integerize
        foo = 0 < Attr(int)               # Positive integer
        foo = (0 <= Attr(int)) <= 200     # Alas, parens are necessary!

        # Stringify, then strip whitespace
        foo = Attr(str).strip()

        # Python 3: Accept bytes or str, decoding if possible (only decodes
        # bytes since decode not a method of str)
        foo = Attr(isa=(bytes, str)).decode("UTF-8")

        # Coerce to a list via .split()
        foo = Attr(isa=(list, str)).split()

    Anything based off of Amethyst's :py:class:`Object` class generally will work as well::

        class MyClass(amethyst.core.Object):
            ...

        class MyClass2(amethyst.core.Object):
            foo = Attr(MyClass)

    :ivar name: Attribute name when assigned to an Object (auto-set by metaclass).
    """
    def __init__(self, convert=None, verify=None, isa=None, default=None, builder=None, fget=None, fset=None, fdel=None, doc=None, OVERRIDE=False):
        """
        :param convert: Attribute converter. Must be a callable or else a
           text string of a class or function name. Classes and functions
           may be imported from other packages by prefixing prepending the
           package name and a dot. For instance, `numpy.array` (see
           :py:func:`amethyst.core.util.get_class` for string processing
           details). Callable should accept a single argument, the value,
           and should return a canonicalized value. Invalid values should
           raise a ValueError(). If converter is `None`, values will be
           accepted unmodified.

        :param isa: Called after conversion but before verification,
           ensures that the value is one of the passed types. Is a shortcut
           for `verify=lambda val: isinstance(val, isa)`

        :param verify: Attribute verifier. Called after conversion, this
           callable should return a truthy result if the value is acceptable.

        :param default: Default value applied at object creation time. If
           default is a callable, it will be called to produce the default
           (e.g., `list`).

            .. note::
              The default value (or result of callable) is assumed valid
              and will not pass through conversion or verification.

        :param builder: Callable which will lazily build a default value
           when the attribute is first used.

        :param fget:
        :param fset:
        :param fdel: If any of fget, fset, or fdel are defined, they will
           be used to construct the object property. If all three are none
           (the default), then the functions which get/set/del the
           appropriate key in the object dictionary will be defined.

        :param doc: Documentation to be attached to the property.

        :param OVERRIDE: When true, allow attribute to replace an existing
           attribute (from a parent class).

        """
        if not (verify is None or callable(verify)):
            raise TypeError("Unknown 'verify' type")
        if not (convert is None or callable(convert) or isinstance(convert, six.text_type)):
            raise TypeError("Unknown 'convert' type")
        self.convert = convert
        self.isa     = isa
        self.verify  = verify
        self.fget    = fget
        self.fset    = fset
        self.fdel    = fdel
        self.doc     = doc
        self.default = default
        self.builder = builder
        self.OVERRIDE = OVERRIDE

        self._package = None
        if isinstance(convert, six.text_type) and (convert.startswith('.') or '.' not in convert):
            try:
                self._package = inspect.getmodule(inspect.stack()[1][0])
            except Exception:
                pass


    def build_property(self, name):
        """ """
        if self.fget is None and self.fset is None and self.fdel is None:

            def fget(obj):
                try:
                    return obj.dict[name]
                except KeyError:
                    # default happens before builder
                    if self.default is not None:
                        obj.dict[name] = self(self.get_default(), name)
                        return obj.dict[name]
                    if self.builder is not None:
                        obj.dict[name] = self(self.builder(), name)
                        return obj.dict[name]
                return None

            def fset(obj, value):
                obj.amethyst_assert_mutable()
                obj.dict[name] = self(value, name)

            def fdel(obj):
                obj.amethyst_assert_mutable()
                del obj.dict[name]

            return property(fget, fset, fdel, self.doc)

        else:
            return property(self.fget, self.fset, self.fdel, self.doc)

    def get_default(self):
        """ """
        if callable(self.default):
            return self.default()
        else:
            return self.default

    def copy_meta(self, *others):
        """
        Copy metadata from another Attr object. This method is used when
        defining derived attributes (e.g., :py:meth:`strip`) to copy
        documentation and the OVERRIDE flag. Returns the object itself for
        chaining.
        """
        for attr in others:
            if (not self.OVERRIDE) and hasattr(attr, "OVERRIDE"): self.OVERRIDE = attr.OVERRIDE
            if (not self.doc)      and hasattr(attr, "doc"):      self.doc = attr.doc
        return self

    def __call__(self, value, key=None):
        """ """
        if self.convert:
            if getattr(self, "_last_convert", None) is not self.convert:
                self._last_convert = self._convert = self.convert
                if not callable(self.convert):
                    self._convert = get_class(self.convert, package=self._package, frame=None)
                self._convert_is_type = isinstance(self._convert, type)
            if not (self._convert_is_type and isinstance(value, self._convert)):
                value = self._convert(value)

        if self.isa:
            if not isinstance(value, self.isa):
                raise ValueError("Value of '{}' is not an instance of {}".format(key, str(self.isa)))

        if self.verify:
            if not self.verify(value):
                raise ValueError("Value of '{}' does not satisfy verification callback".format(key))

        return value

    def __and__(self, other):
        """ """
        return self.__class__(lambda v: other(self(v))).copy_meta(self, other)
    def __rand__(self, other):
        """ """
        return self.__class__(lambda v: self(other(v))).copy_meta(self, other)

    def __or__(self, other):
        """ """
        def convert(value):
            try:
                return self(value)
            except ValueError as err:
                if callable(other):
                    return other(value)
                elif value == other:
                    return other
                else:
                    raise err
        return self.__class__(convert).copy_meta(self, other)

    def __ror__(self, other):
        """ """
        def convert(value):
            try:
                if callable(other):
                    return other(value)
                elif value == other:
                    return other
                else:
                    raise ValueError("Invalid value")
            except ValueError:
                return self(value)
        return self.__class__(convert).copy_meta(self, other)

    __hash__ = None   # has __eq__ but isn't hashable
    def __eq__(self, other):
        """
        Tests via :py:func:`amethyst.core.util.smartmatch`

        .. WARNING::
           Hash lookups must be idempotent (looking up the result of
           a previous lookup had better return the same thing) since we offer
           no guarantees that validation may not happen more than once.

           GOOD:: :code:`{ "a": "A", "b": "B",  "A": "A", "B": "B" }`

           BAD:: :code:`{ "a": "A", "b": "B" }  # will fail on repeated validation since "A" and "B" are not keys`
        """
        return self.__class__(lambda v: smartmatch(self(v), other)).copy_meta(self)
    def __ne__(self, other):
        """Ensure no smartmatch"""
        def convert(value):
            val = self(value)
            try:
                # If we do not match the other value, this raises an
                # exception (thus we can return val).
                smartmatch(val, other)
            except ValueError:
                return val
            # otherwise, we've matched the smartmatch, this we match what
            # we don't want to be - raise a value error.
            raise ValueError("Invalid Value")
        return self.__class__(convert).copy_meta(self)

    # This is starting to get cute:
    def __lt__(self, other):
        """ """
        def convert(value):
            val = self(value)
            if val < other: return val
            raise ValueError("Invalid Value")
        return self.__class__(convert).copy_meta(self)
    def __le__(self, other):
        """ """
        def convert(value):
            val = self(value)
            if val <= other: return val
            raise ValueError("Invalid Value")
        return self.__class__(convert).copy_meta(self)
    def __ge__(self, other):
        """ """
        def convert(value):
            val = self(value)
            if val >= other: return val
            raise ValueError("Invalid Value")
        return self.__class__(convert).copy_meta(self)
    def __gt__(self, other):
        """ """
        def convert(value):
            val = self(value)
            if val > other: return val
            raise ValueError("Invalid Value")
        return self.__class__(convert).copy_meta(self)

    # These modifiers make no sense unless they are idempotent since we may
    # validate multiple times. Thus, we only define those whose semantics
    # swing that way.
    def __mod__(self, other):
        """ """
        return self.__class__(lambda v: self(v) % other).copy_meta(self)
    def __pos__(self):
        """ """
        return self.__class__(lambda v: +self(v)).copy_meta(self)
    def __abs__(self):
        """ """
        return self.__class__(lambda v: abs(self(v))).copy_meta(self)

    # I don't see much use for float() since it is the first thing you
    # would want to do. However, int() could be useful since
    # Attr(float).int() is a more flexible converter (integerizing stringy
    # floats rather than raising an exception). Just for completeness, we
    # include complex too.
    def float(self):
        """ """
        return self.__class__(lambda v: float(self(v))).copy_meta(self)
    def int(self):
        """ """
        return self.__class__(lambda v: int(self(v))).copy_meta(self)
    def complex(self):
        """ """
        return self.__class__(lambda v: complex(self(v))).copy_meta(self)

    # Can also define a handful of common methods one might wish to call,
    # and call them if present. Happy duck-typing.
    def strip(self, chars=None):
        """Return a new attribute which strips whitespace if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.strip(chars) if hasattr(value, "strip") else value
        return self.__class__(convert).copy_meta(self)
    def rstrip(self, chars=None):
        """Return a new attribute which strips whitespace from the right side if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.rstrip(chars) if hasattr(value, "rstrip") else value
        return self.__class__(convert).copy_meta(self)
    def lstrip(self, chars=None):
        """Return a new attribute which strips whitespace from the left side if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.lstrip(chars) if hasattr(value, "lstrip") else value
        return self.__class__(convert).copy_meta(self)

    def encode(self, encoding="UTF-8", errors="strict"):
        """Return a new attribute which encodes value if applicable (duck typing). Defaults to UTF-8 encoding."""
        def convert(value):
            value = self(value)
            return value.encode(encoding, errors) if hasattr(value, "encode") else value
        return self.__class__(convert).copy_meta(self)
    def decode(self, encoding="UTF-8", errors="strict"):
        """Return a new attribute which decodes value if applicable (duck typing). Defaults to UTF-8 encoding."""
        def convert(value):
            value = self(value)
            return value.decode(encoding, errors) if hasattr(value, "decode") else value
        return self.__class__(convert).copy_meta(self)

    def lower(self):
        """Return a new attribute which lower-cases value if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.lower() if hasattr(value, "lower") else value
        return self.__class__(convert).copy_meta(self)
    def upper(self):
        """Return a new attribute which upper-cases value if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.upper() if hasattr(value, "upper") else value
        return self.__class__(convert).copy_meta(self)
    def title(self):
        """Return a new attribute which title-cases value if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.title() if hasattr(value, "title") else value
        return self.__class__(convert).copy_meta(self)
    def capitalize(self):
        """Return a new attribute which capitalizes value if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.capitalize() if hasattr(value, "capitalize") else value
        return self.__class__(convert).copy_meta(self)
    def casefold(self):
        """Return a new attribute which casefolds value if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.casefold() if hasattr(value, "casefold") else value
        return self.__class__(convert).copy_meta(self)

    def split(self, sep=None, maxsplit=-1):
        """Return a new attribute which splits its value if applicable (duck typing)."""
        def convert(value):
            value = self(value)
            return value.split(sep, maxsplit) if hasattr(value, "split") else value
        return self.__class__(convert).copy_meta(self)


global_amethyst_encoders = dict()
global_amethyst_hooks = dict()
def register_amethyst_type(cls, encode, decode, name=None, overwrite=False, wrap_encode=True):
    """
    Adds a type to the global list (:py:data:`global_amethyst_encoders`)
    for object encoding and decoding. Subclasses of :py:class:`Object` are
    automatically registered, so you should only need to register external
    objects that you use.

    :param class cls: Class to register

    :param callable encode: Callable which will transform the object into a "dumb"
        structure of primitive objects (dict, list, str, int, float).

    :param callable decode: Callable which will transform the "dumb" structure of
        primitive objects back into the object.

    :param str name: Globally unique string identifying the class. This
        name should be something which won't appear as a dictionary key in
        normal data. Defaults to "__MODULENAME.CLASSNAME__"

    :param bool overwrite: By default, this function will raise an error if
        a class or name is aready registered. Pass True to override any
        existing registrations.

    :param bool wrap_encode: By default, the encoded object will be wrapped
        in a single-key dict: :code:`{ name: ENCODED_OBJECT }` so that the
        decoder can be called when inflating a structure containing the
        object. Pass True in this parameter in order to avoid wrapping the
        encoded structure.

        |br|

        You might want to set this parameter if your object is naturally
        expressed as a basic object and you are certain that all uses after
        inflation will automatically coerce the value to your desired
        object when needed. For instance, a URL object where all functions
        and methods which accept the URL object also accept a plain string.
        You could pass an encoder which deflates to plain strings and set
        :code:`wrap_encode=False` and then URLs would appear as plain strings in
        your exported structures, which may be easier to work with in
        external applications.

        |br|

        This option also offers an escape hatch for hypothetical cases
        where you may need to wrap your encoded object in your encoder
        itself.

    .. seealso:: :py:func:`amethyst_deflate` and :py:func:`amethyst_inflate`

    """
    if name is None:
        if isinstance(cls, BaseObject):
            name = cls._dundername
        else:
            name = "__{}.{}__".format(cls.__module__, cls.__name__)
    if cls in global_amethyst_encoders and not overwrite:
        raise ValueError("Class encoder '{}' already reqistered".format(cls))
    if name in global_amethyst_hooks and not overwrite:
        raise ValueError("Class hook '{}' already reqistered".format(name))
    if wrap_encode:
        global_amethyst_encoders[cls] = lambda obj: { name: encode(obj) }
    else:
        global_amethyst_encoders[cls] = encode
    global_amethyst_hooks[name] = decode


def amethyst_deflate(obj, deflator=None):
    """
    Deflate a structure of amethyst-encodable objects into a "dumb"
    structure of plain dicts, lists, numbers, and strings. The deflated
    structure should be easily serializable by most any reasonable
    serialization library (yaml, lxml, ...)

    Makes use of :py:data:`global_amethyst_encoders` by default. Pass an amethyst
    Object as second argument to make use of any Object-local encoders.

    Note: If your target is JSON, the amethyst object's :py:func:`Object.toJSON` method is
    probably better.
    """
    global global_amethyst_encoders

    if obj is None or isinstance(obj, (six.string_types, six.binary_type, numbers.Number, bool)):
        return obj
    elif isinstance(obj, dict):
        return { six.text_type(k): amethyst_deflate(obj[k], deflator) for k in obj }
    elif isinstance(obj, list):
        return [ amethyst_deflate(k, deflator) for k in obj ]
    elif isinstance(obj, tuple):
        return tuple(amethyst_deflate(k, deflator) for k in obj)
    elif hasattr(deflator, "_jsonencoders") and obj.__class__ in deflator._jsonencoders:
        return amethyst_deflate(deflator._jsonencoders[obj.__class__](obj), deflator)
    elif obj.__class__ in global_amethyst_encoders:
        return amethyst_deflate(global_amethyst_encoders[obj.__class__](obj), deflator)
    raise TypeError("Can't encode {}".format(repr(obj)))


def amethyst_inflate(obj, inflator=None, start=0):
    """
    Inflate an annotated "dumb" structure to a structure of objects, the
    opposite of :py:func:`amethyst_deflate`.

    This function should be used to assist in inflation of ``Object``
    instances using alternative serialization tools. Any serializer that
    can produce plain dicts and lists should work with this function.

    This function handles either *single-key* or *flat* style annotations.
    Un-annotated constructions will pass through unchanged.

    This function makes use of :py:data:`global_amethyst_encoders` by
    default. Pass an amethyst ``Object`` as second argument to make use of
    any Object-local encoders.

    .. note::
      If your source is JSON, the amethyst object's :py:func:`Object.fromJSON`
      or class :py:func:`Object.newFromJSON` method is probably better.
    """
    global global_amethyst_hooks

    if isinstance(obj, dict):
        # Look for *single-key* or *flat* style annotations
        key = data = None
        if start <= 0:
            if '__class__' in obj:
                data = dict(obj)
                key = amethyst_inflate(data.pop('__class__'), inflator, start=1)
            elif 1 == len(obj):
                key = list(obj.keys())[0]
                data = amethyst_inflate(obj[key], inflator, start=1)

        # If annotated, try to inflate
        if key is not None:
            if hasattr(inflator, "_jsonhooks") and key in inflator._jsonhooks:
                return inflator._jsonhooks[key](data)
            elif key in global_amethyst_hooks:
                return global_amethyst_hooks[key](data)

        # If not annotated or inflation fails (unknown class), inflate values:
        for key in obj:
            obj[key] = amethyst_inflate(obj[key], inflator, start=(start-1))

    elif isinstance(obj, list):
        for i, val in enumerate(obj):
            obj[i] = amethyst_inflate(val, inflator, start=(start-1))

    return obj


# Python3 moved the builtin modules around, force the name so py3 can talk to py2
register_amethyst_type(set, list, set, name="__set__")
register_amethyst_type(frozenset, list, frozenset, name="__frozenset__")


class AttrsMetaclass(type):
    """
    Metaclass for Amethyst Object class descendants. Simply looks at all
    attributes for any which are instances of :py:class:`Attr`. The
    :py:class:`Attr` itself is saved to the :py:attr:`_attrs` class
    attribute (a dictionary) and a property created in its place via the
    Attr :py:func:`Attr.build_property` method.
    """
    def __new__(cls, class_name, bases, attrs):
        new_attrs = dict()
        for name in list(attrs.keys()):
            if isinstance(attrs[name], Attr):
                new_attrs[name] = attrs.pop(name)
                new_attrs[name].name = name

        new_cls = super(AttrsMetaclass, cls).__new__(cls, class_name, bases, attrs)

        # Need some bootstrapping
        if class_name != 'BaseObject' and attrs.get("amethyst_register_type", True):
            register_amethyst_type(
                new_cls,
                encode    = (lambda obj: obj.dict),
                decode    = (lambda obj: new_cls(obj)),
                overwrite = False
            )

        # Merge json hooks into the base _json* hooks.
        for jattr in "jsonhooks", "jsonencoders":
            _jattr = "_" + jattr
            setattr(new_cls, _jattr, dict(getattr(new_cls, _jattr, ())))# Shallow clone
            if hasattr(new_cls, jattr):
                getattr(new_cls, _jattr).update(getattr(new_cls, jattr))
                delattr(new_cls, jattr)

        for name, attr in six.iteritems(new_attrs):
            if not attr.OVERRIDE and hasattr(new_cls, name):
                raise DuplicateAttributeException("Attribute {} in {} already defined in a parent class.".format(name, cls.__name__))
            setattr(new_cls, name, attr.build_property(name))

        new_cls._attrs = new_cls._attrs.copy() if hasattr(new_cls, '_attrs') else dict()
        new_cls._attrs.update(new_attrs)
        new_cls._dundername = "__{}.{}__".format(new_cls.__module__, new_cls.__name__)

        return new_cls


# Manually create a base object so that we can run in both python 2 and 3.
#
#   https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef#metaclasses
BaseObject = AttrsMetaclass(str('BaseObject'), (), {})

class Object(BaseObject):
    """
    Amethyst Base Object

    :ivar _attrs: Dictionary mapping attribute names to :py:class:`Attr`
      objects. Should not be modified, but can be read for introspection of
      an Object.

    :ivar _jsonencoders: Dictionary mapping class objects to callable
      encoders which should produce a JSON-serializable object. These
      functions are called from the JSONEncoder method. Per the json
      documentation, these functions should return an object which is JSON
      serializable or else raise a TypeError. These encoders are specific
      to the class. Use :py:func:`register_amethyst_type` to register a
      class globally.

      .. note::
        _jsonencoders is a lower-level tool than
        :py:func:`register_amethyst_type` and offers direct access to the
        encoders (behaves like `overwrite=True, wrap_encode=False`)

    :ivar _jsonhooks: Dictionary mapping class identifiers (strings of form
      "__MODULENAME.CLASSNAME__") to callable decoders which should inflate
      simple structures to corresponding objects. These functions are
      called from the JSONObjectHook method when inflating data. These
      decoders are specific to the class. Use
      :py:func:`register_amethyst_type` to register a class globally.

      .. note::
        _jsonhooks is a lower-level tool than
        :py:func:`register_amethyst_type` and offers direct access to the
        decoders (behaves like `overwrite=True, wrap_encode=False`)

    """

    amethyst_includeclass  = True
    """
    When True (the default), serialization will include a key "__class__"
    containing the class name of the object which can be used during
    loading to verify that the object is of the correct type.
    """

    amethyst_verifyclass   = True
    """
    When True (the default), loading data from JSON or a dict passed to
    :py:func:`amethyst_load_data()` will check for the "__class__" key described above, and
    an exception will be thrown if it is not found.
    """

    amethyst_import_strategy = "strict"
    """
    When "strict" (the default), then loading data from JSON or a
    dictionary via :py:func:`amethyst_load_data()` requires all keys present in the data
    structure to correspond with keys in the attribute list. If any
    additional keys are present, an exception will be raised. When "loose",
    additional keys will be ignored and not copied into the object
    dictionary. When "sloppy", unknown attributes will be copied unmodified
    into the object dict.
    """

    amethyst_register_type = True
    amethyst_classhint_style = "flat"

    def __init__(self, *args, **kwargs):
        """
        Initializes self.dict with all passed kwargs.
        Object is mutable by default.

        .. warning::
           Passing a single argument that is an instance of the class
           itself is reserved for internal use only and behavior may
           change.
        """
        super(Object, self).__init__()
        self._amethyst_mutable_ = True

        # Special-case of single argument:
        if len(args) == 1 and not kwargs and isinstance(args[0], type(self)):
            warnings.warn("amethyst shallow-copy, this use case may become deprecated. If you have a use for it, notify upstream developer soon!", DeprecationWarning)
            # - NOT A COPY! Primary case of Object(Object()) is
            #   re-import by JSON Hook, thus we just need another view
            #   of the same data (returning the same object would also
            #   work, but I don't know how to do that, and isn't really
            #   worth it).
            #
            # - Object already one of our type, no need to merge defaults
            self.dict = args[0].dict
            self._amethyst_mutable_ = args[0]._amethyst_mutable_

        else:
            self.dict = dict()
            data = dict()
            for d in args: data.update(d)
            if kwargs: data.update(kwargs)
            if data:
                self.amethyst_load_data(data, verifyclass=False)

        for name, attr in six.iteritems(self._attrs):
            if attr.default is not None and name not in self.dict:
                self.dict[name] = attr.get_default()

    def amethyst_assert_mutable(self, msg="May not modify, object is immutable"):
        """ """
        if not self._amethyst_mutable_:
            raise ImmutableObjectException(msg)
        return self

    def amethyst_is_mutable(self):
        """ """
        return self._amethyst_mutable_
    def amethyst_make_mutable(self):
        """ """
        self._amethyst_mutable_ = True
        return self
    def amethyst_make_immutable(self):
        """ """
        self._amethyst_mutable_ = False
        return self

    def __str__(self):
        return str(self.dict)
    def __repr__(self):
        return repr(self.dict)

    def __len__(self):
        """ """
        return len(self.dict)
    def __contains__(self, key):
        """ """
        return key in self.dict
    def __iter__(self):
        """ """
        return iter(self.dict)

    __hash__ = None   # has __eq__ but isn't hashable
    def __eq__(self, other):
        """
        Object equality. Tests all :py:func:`Attr()`s and only
        :py:func:`Attr()`s. Other python properties or "garbage" in the
        underlying dict (which may arise from "sloppy" imports) will be
        ignored.
        """
        if self.__class__ != other.__class__:
            return False
        for name, attr in six.iteritems(self._attrs):
            if getattr(self, name, UNIQUE1) != getattr(other, name, UNIQUE2):
                return False
        return True
    def __ne__(self, other):
        return not (self == other)

    def items(self, **kwargs):
        """
        Subclasses: this method may be overridden with an unrelated implementation.
        """
        for name, attr in six.iteritems(self._attrs):
            if name in self.dict:
                yield name, self.dict[name]
    iteritems = items
    def keys(self, **kwargs):
        """
        Subclasses: this method may be overridden with an unrelated implementation.
        """
        for name, attr in six.iteritems(self._attrs):
            if name in self.dict:
                yield name
    def values(self, **kwargs):
        """
        Subclasses: this method may be overridden with an unrelated implementation.
        """
        for name, attr in six.iteritems(self._attrs):
            if name in self.dict:
                yield self.dict[name]

    def __getitem__(self, key):
        """ """
        return self.dict[key]
    def __setitem__(self, key, value):
        """ """
        self.amethyst_assert_mutable()
        attr = self._attrs.get(key)
        if attr is not None:
            self.dict[key] = attr(value, key)
        elif self.amethyst_import_strategy == "strict":
            raise KeyError("key {} not permitted in {} object".format(key, self._dundername))
        elif self.amethyst_import_strategy == "sloppy":
            self.dict[key] = value
    def __delitem__(self, key):
        """ """
        self.amethyst_assert_mutable()
        del self.dict[key]

    def get(self, key, dflt=None):
        """
        Subclasses: this method may be overridden with an unrelated implementation.
        """
        return self.dict.get(key, dflt)

    def set(self, *args, **kwargs):
        """
        Verify then set canonicalized value. Positional args take
        precedence over kwargs. ::

            obj.set(key, val)
            obj.set(foo=val)

        Subclasses: this method may be overridden with an unrelated implementation.
        """
        if 2 == len(args) and not kwargs:
            self[args[0]] = args[1]
            return self
        self.amethyst_assert_mutable()
        for i in range(0, len(args), 2):
            kwargs[args[i]] = args[i+1]
        self.dict.update(self.amethyst_validate_update(kwargs))
        return self

    def setdefault(self, key, value):
        """
        If missing a value, verify then set

        Subclasses: this method may be overridden with an unrelated implementation.
        """
        self.amethyst_assert_mutable()
        if key not in self.dict:
            self.set(key, value)
        return self.dict[key]

    def direct_set(self, *args, **kwargs):
        """
        Set values BYPASSING VALIDATION but respecting mutability.
        Positional args take precedence over kwargs. ::

            obj.direct_set(key, val)
            obj.direct_set(foo=val)

        Subclasses: this method may be overridden with an unrelated implementation.
        """
        self.amethyst_assert_mutable()
        for i in range(0, len(args), 2):
            kwargs[args[i]] = args[i+1]
        self.dict.update(kwargs)
        return self

    def pop(self, key, dflt=None):
        """
        Subclasses: this method may be overridden with an unrelated implementation.
        """
        self.amethyst_assert_mutable()
        return self.dict.pop(key, dflt)

    def update(self, *args, **kwargs):
        """
        Subclasses: this method may be overridden with an unrelated implementation.
        """
        self.amethyst_assert_mutable()
        data = dict()
        for d in args: data.update(d)
        if kwargs: data.update(kwargs)
        self.dict.update(self.amethyst_validate_update(data))
        return self

    def direct_update(self, *args, **kwargs):
        """
        Update internal dictionary BYPASSING VALIDATION but respecting
        mutability.

        Subclasses: this method may be overridden with an unrelated implementation.
        """
        self.amethyst_assert_mutable()
        self.dict.update(*args, **kwargs)
        return self

    def amethyst_validate_update(self, d, import_strategy=None):
        """
        Convert and validate with the intention of updating only some of
        the object's .dict values. Returns a new dictionary with
        canonicalized values, but *does not initialize any missing keys
        with attribute default values* (which distinguishes this from
        :py:func:`amethyst_validate_data`).

        This method does not change the object. Pass the resulting dict to
        the `.update()` method (or `.direct_update()` if you decide to accept
        the changes.
        """
        strategy = coalesce(import_strategy, self.amethyst_import_strategy)
        data = d.copy() if strategy == "sloppy" else dict()
        for key, val in six.iteritems(d):
            attr = self._attrs.get(key)
            if attr is None and strategy == "strict":
                raise KeyError("key {} not permitted in {} object".format(key, self._dundername))
            elif attr is not None:
                data[key] = attr(val, key)
        return data

    def amethyst_validate_data(self, d, import_strategy=None):
        """
        Convert and validate with the intention of replacing all of the
        object's .dict values. Returns a new dictionary with canonicalized
        values, *and defaults inserted* (which distinguishes this from
        :py:func:`amethyst_validate_update`).

        This method does not change the object. Typical usage would look
        like either::

            myobj.dict = myobj.amethyst_validate_data(data)

        or ::

            validated = myobj.amethyst_validate_data(data)
            mynewobj = MyClass(**validated)

        Subclasses of :py:class:`Object` can also use this method to
        inflate specific attibutes at load time. For instance, to inflate
        non-Object objects or ensure objects from hand-written config
        files. Be sure to override :py:func:`amethyst_validate_update` as
        well if programmatic updates may need special inflation rules.
        """
        strategy = coalesce(import_strategy, self.amethyst_import_strategy)
        data = d.copy() if strategy == "sloppy" else dict()
        keys = set(d.keys()) if strategy == "strict" else set()
        for name, attr in six.iteritems(self._attrs):
            keys.discard(name)
            if name in d:
                data[name] = attr(d[name], name)
            elif attr.default is not None:
                data[name] = attr.get_default()
        if keys:
            raise ValueError("keys {} not permitted in {} object".format(keys, self._dundername))
        return data

    def attr_value_ok(self, name, value):
        """
        Validate a single value independently of any others. Just checks
        that the attribute validator does not raise an exception.

        Subclasses: this method may be overridden with an unrelated implementation.
        """
        if name in self._attrs:
            try:
                self._attrs[name](value, name)
                return True
            except ValueError:
                return False
        return False

    def amethyst_load_data(self, data, import_strategy=None, verifyclass=None):
        """
        Loads a data dictionary with validation. Modifies the passed dict
        and replaces current self.dict object with the one passed.

        :param import_strategy: Provides a local override to the :py:attr:`amethyst_import_strategy` class attribute.

        :param verifyclass: Provides a local override to the :py:attr:`amethyst_verifyclass` class attribute.

        This method transparently loads data in either "single-key" or "flat" formats::

            { "__my.module.MyClass__": { ... obj.dict ... } }

            { "__class__": "MyClass", ... obj.dict ... }

        Keep in mind that the default base value for :py:attr:`amethyst_verifyclass` is
        True, so, by default, at least one of the class identification keys
        is expected to be present.
        """
        self.amethyst_assert_mutable()
        verifyclass = coalesce(verifyclass, self.amethyst_verifyclass)

        # We only deal in dicts here
        if isinstance(data, self.__class__):
            verifyclass = False

        if isinstance(data, Object):
            data = data.dict

        if not isinstance(data, dict):
            raise ValueError("expected dictionary object")

        # Accept data in single-key mode. Pop out the inner dict and, if
        # the indicated class name is what we expect, we can bypass the
        # class verification step.
        if 1 == len(data):
            for key in data:
                if key == self._dundername:
                    verifyclass = False# We're good
                    data = data[key]
                    if not isinstance(data, dict):
                        raise ValueError("expected dictionary object")

        # verifyclass may need to be locally overridden if the source is
        # broken. Once we do verify the class, remove it from the dict to
        # make key iteration safe.
        if verifyclass and data.get("__class__") != self._dundername:
            if data.get("__class__") is None:
                raise ValueError("Error validating import data class: __class__ key missing but should be {} (or set verifyclass=False)".format(self._dundername))
            else:
                raise ValueError("Error validating import data class: got {} object, but expected {}".format(data.get("__class__"), self._dundername))
        data.pop("__class__", None)

        # Run the validator
        self.dict = self.amethyst_validate_data(data, import_strategy=import_strategy)
        return self
    load_data = amethyst_load_data
    """
    Alias for amethyst_load_data

    Subclasses: this method may be overridden with an unrelated implementation.
    """

    def JSONEncoder(self, obj):
        """
        Fallback method for JSON encoding.

        If the standard JSONEncoder is unable to encode an object, this
        method will be called. Per the json documentation, it should return
        an object which is JSON serializable or else raise a TypeError.

        This base encoder, looks up an object's class in a dict and calls
        the corresponding function to do the translation. The built-in
        translators map::

            set       => { "__set__": [ ... ] }
            frozenset => { "__frozenset__": [ ... ] }

        Additional translators may be added by creating a class variable
        :py:attr:`jsonencoders` which is a dict mapping classes to a function. These
        translators will merged onto the base translators (silently
        replacing duplicates) by the metaclass at class (not object)
        creation.
        """
        global global_amethyst_encoders

        if obj.__class__ in self._jsonencoders:
            return self._jsonencoders[obj.__class__](obj)
        elif obj.__class__ in global_amethyst_encoders:
            return global_amethyst_encoders[obj.__class__](obj)
        raise TypeError("Can't encode {}".format(repr(obj)))

    def JSONObjectHook(self, obj):
        """
        Object hook for JSON decoding.

        This method is called for every decoded JSON object. If necessary,
        it should return a new or modified object that should be used instead.

        This base encoder, translates single-key dicts into new objects if
        the single-key is a special value. The built-in translators are::

            { "__set__": [ ... ] }       => set
            { "__frozenset__": [ ... ] } => frozenset

        Additional translators may be added by creating a class variable
        :py:attr:`jsonhooks` which is a dict mapping the special key to a function.
        These translators will merged onto the base translators (silently
        replacing duplicates) by the metaclass at class (not object)
        creation.

        Keep in mind that JSON input comes from untrusted sources, so
        translators will need to be robust against malformed structures.
        """
        global global_amethyst_hooks
        if isinstance(obj, dict) and 1 == len(obj):
            for key in obj:
                if key in self._jsonhooks:
                    return self._jsonhooks[key](obj[key])
                elif key in global_amethyst_hooks:
                    return global_amethyst_hooks[key](obj[key])
        return obj

    def toJSON(self, includeclass=None, style=None, **kwargs):
        """
        Paramters are sent directly to json.dumps except:

        :param includeclass: When true, include a class indicator using the
           method requested by the :code:`style` parameter. When `None` (the
           default), defer to the value of the class variable
           :py:attr:`amethyst_includeclass`.

        :param style: When including class, what style to use (root-level
           object only). Options are:

             * "flat" to produce a JSON string in the form::

                { "__class__": "__my.module.MyClass__", ... obj.dict ... }

             * "single-key" to produce a JSON string in the form::

                { "__my.module.MyClass__": { ... obj.dict ... } }

        The default style is taken from the class :py:attr:`amethyst_classhint_style`
        attribute.
        """
        kwargs.setdefault('default', self.JSONEncoder)
        includeclass = coalesce(includeclass, self.amethyst_includeclass)
        style = coalesce(style, self.amethyst_classhint_style)
        popclass = False

        try:
            dump = self.dict
            if includeclass:
                if style == "flat":
                    popclass = True
                    self.dict["__class__"] = self._dundername
                elif style == "single-key":
                    dump = { self._dundername: self.dict }
                else:
                    raise AmethystException("Unknown class style '{}'".format(style))
            rv = json.dumps(dump, **kwargs)
        finally:
            if popclass:
                self.dict.pop("__class__", None)
        return rv

    @classmethod
    def newFromJSON(cls, source, import_strategy=None, verifyclass=None, **kwargs):
        """ """
        self = cls()
        mutable = self.amethyst_is_mutable()# In case some subclass is default immutable
        if not mutable: self.amethyst_make_mutable()
        self.fromJSON(source, import_strategy=import_strategy, verifyclass=verifyclass, **kwargs)
        if not mutable: self.amethyst_make_immutable()
        return self

    def fromJSON(self, source, import_strategy=None, verifyclass=None, **kwargs):
        """
        Paramters are sent directly to json.load or json.loads except:

        :param import_strategy: Provides a local override to the :py:attr:`amethyst_import_strategy` class attribute.

        :param verifyclass: Provides a local override to the :py:attr:`amethyst_verifyclass` class attribute.
        """
        kwargs.setdefault('object_hook', self.JSONObjectHook)
        if isinstance(source, six.string_types):
            data = json.loads(source, **kwargs)
        else:
            data = json.load(source, **kwargs)
        return self.amethyst_load_data(data, import_strategy=import_strategy, verifyclass=verifyclass)

    def deflate_data(self):
        """
        Deflate object into a "dumb" structure of plain dicts, lists,
        numbers, and strings. The deflated structure should be easily
        serializable by most any reasonable serialization library (yaml,
        lxml, ...)

        Subclasses: this method may be overridden with an unrelated implementation.
        """
        return amethyst_deflate(self.dict, self)

    @classmethod
    def inflate_new(cls, obj):
        """
        Inflate a "dumb" structure to a structure of objects, the opposite
        of :py:func:`deflate_data()`. Allows inflation from arbitrary serialization
        tools, as long as they can produce dicts and lists.

        Subclasses: this method may be overridden with an unrelated implementation.
        """
        rv = amethyst_inflate(obj, cls)
        return rv if rv.__class__ is cls else cls(rv)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: LGPL-3.0

from __future__ import division, absolute_import, print_function, unicode_literals
import unittest

import json
import marshal
import six
import sqlite3

import amethyst.core.obj
from amethyst.core import ImmutableObjectException, DuplicateAttributeException
from amethyst.core import Object, Attr
from amethyst.core import amethyst_deflate, amethyst_inflate


class Obj(Object):
    foo = Attr(int)
    bar = Attr(isa=six.text_type).strip()
    baz = Attr(float)
    bip = Attr()


class MyTest(unittest.TestCase):
    def test_ttobject(self):
        obj = Object()
        self.assertEqual(obj.dict, {}, "Initial object is empty")
        self.assertEqual(len(obj), 0, "Empty length works")

        obj = Object({})
        self.assertEqual(obj.dict, {}, "Initial object is empty")

        self.assertEqual(json.loads(obj.toJSON()), {"__class__": "__amethyst.core.obj.Object__"}, "Empty toJSON")

        obj = Object.newFromJSON('{"__amethyst.core.obj.Object__": {}}')
        obj = Object.newFromJSON('{"__class__": "__amethyst.core.obj.Object__"}')

        self.assertIsNone(obj.get("__class__"), "fromJSON removes __class__ key")

        with self.assertRaises(ValueError, msg="class verification works"):
            obj = Object.newFromJSON('{"__amethyst.core.obj.NotObject__": {}}')

        with self.assertRaises(ValueError, msg="Invalid if __class__ is missing"):
            obj.fromJSON('{}')

        self.assertFalse(amethyst.core.obj.UNIQUE1 == amethyst.core.obj.UNIQUE2)
        self.assertTrue( amethyst.core.obj.UNIQUE1 != amethyst.core.obj.UNIQUE2)

        class ObjA(Object):
            foo = Attr(int)
            bar = Attr(isa=six.text_type).strip()

        class ObjB(Object):
            foo = Attr(int)
            bar = Attr(isa=six.text_type).strip()

        class ObjC(ObjA):
            pass

        a = ObjA(foo=23, bar="plugh")
        b = ObjB(foo=23, bar="plugh")
        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue( b != a)
        self.assertTrue( a != b)

        a = ObjA(foo=23, bar="plugh")
        c = ObjC(foo=23, bar="plugh")
        self.assertFalse(a == c)
        self.assertFalse(c == a)
        self.assertTrue( c != a)
        self.assertTrue( a != c)

        a1 = ObjA(foo=23, bar="plugh")
        a2 = ObjA(foo=23, bar="plugh")
        self.assertTrue( a1 == a2)
        self.assertFalse(a1 != a2)

        a1 = ObjA(foo=23, bar="plugh")
        a2 = ObjA(foo=23, bar="barf")
        self.assertFalse(a1 == a2)

        a1 = ObjA(foo=23, bar="plugh")
        a2 = ObjA(bar="plugh")
        self.assertFalse(a1 == a2)

        a1 = ObjA(foo=23, bar="plugh")
        a2 = ObjA()
        self.assertFalse(a1 == a2)

        a1 = ObjA()
        a2 = ObjA(foo=23, bar="plugh")
        self.assertFalse(a1 == a2)

        a1 = ObjA()
        a2 = ObjA()
        self.assertTrue(a1 == a2)

        a1 = ObjA(foo=24)
        a2 = ObjA(foo=24)
        self.assertTrue(a1 == a2)
        self.assertTrue(a1 is not a2)

        self.assertEqual(a1.pop("foo"), 24)
        self.assertEqual(list(a1.keys()), [])

        self.assertTrue(a1.attr_value_ok("foo", 12))
        self.assertTrue(a1.attr_value_ok("foo", "23"))
        self.assertTrue(not a1.attr_value_ok("foo", "2.3"))
        self.assertTrue(not a1.attr_value_ok("plugh", 23))


    def test_overrides(self):
        """
        Some of our methods are explicitly allowed to be replaced. These
        tests replace them with exceptions and verify that other method
        calls don't accidentally call the overrides.
        """
        class ObjOver1(Object):
            foo = Attr(int)
            bar = Attr(isa=six.text_type).strip()

            def croak(self, *args, **kwargs):
                raise Exception("Bummer")

            items = croak
            keys = croak
            values = croak
            get = croak
            set = croak
            setdefault = croak
            direct_set = croak
            pop = croak
            update = croak
            direct_update = croak
            attr_value_ok = croak
            load_data = croak
            deflate_data = croak
            inflate_new = croak

        myobj = ObjOver1(foo=23)

        with six.assertRaisesRegex(self, Exception, r'Bummer'):
            myobj.items()

        myobj.other = 15
        self.assertEqual(myobj.other, 15)
        self.assertEqual(myobj.bar, None)
        self.assertEqual(len(myobj), 1)
        self.assertTrue("foo" in myobj)
        self.assertEqual(myobj.foo, 23)
        self.assertEqual(myobj['foo'], 23)
        del myobj['foo']
        self.assertEqual(myobj.foo, None)
        myobj.foo = 24
        self.assertEqual(myobj.foo, 24)
        myobj['foo'] = 25
        self.assertEqual(myobj.foo, 25)
        myobj.amethyst_validate_update(dict(foo=26))
        myobj.amethyst_validate_data(dict(foo=27))
        myobj.amethyst_load_data(dict(foo=28), verifyclass=False)
        self.assertEqual(myobj.foo, 28)
        self.assertNotEqual(myobj.toJSON(), '')
        myobj.fromJSON('{"foo":"29"}', verifyclass=False)
        self.assertEqual(myobj.foo, 29)


    def test_other_attrs(self):
        """Other attributes explicitly allowed and are not stored in data hash"""
        myobj = Obj(foo=23)
        myobj.other = 15
        self.assertEqual(myobj.other, 15)
        self.assertEqual(myobj.deflate_data().get("other", None), None)
        self.assertFalse("other" in myobj.toJSON())

        # other not listed in keys or values
        self.assertEqual(list(myobj.keys()), ["foo"])
        self.assertEqual(list(myobj.values()), [23])

        with self.assertRaises(AttributeError, msg="Getattr raises exception for names undeclared attrs"):
            myobj.undefined
        with self.assertRaises(KeyError, msg="Getitem raises exception for names undeclared attrs"):
            myobj['undefined']
        with self.assertRaises(KeyError):
            myobj['undefined'] = 42
        with self.assertRaises(KeyError):
            print(myobj['undefined'])
        self.assertEqual(myobj.deflate_data().get("undefined", None), None)
        self.assertFalse("undefined" in myobj.toJSON())

        # However, the dict interface should be for proper attrs only
        with self.assertRaises(KeyError):
            print(myobj['other'])
        with self.assertRaises(KeyError):
            myobj['other'] = 42
        self.assertEqual(myobj.other, 15)
        self.assertEqual(myobj.deflate_data().get("other", None), None)
        self.assertFalse("other" in myobj.toJSON())


    def test_README_validation(self):
        class MyObject(Object):
            amethyst_verifyclass = False
            amethyst_register_type = False
            foo = Attr(int)
            bar = Attr(isa=six.text_type).strip()

        # constructors
        myobj = MyObject({ "foo": "23", "bar": "Hello " })
        self.assertIsInstance(myobj.foo, int)
        self.assertEqual(myobj.foo, 23)
        self.assertEqual(myobj.bar, "Hello")

        myobj = MyObject(foo="23", bar="Hello ")
        self.assertIsInstance(myobj.foo, int)
        self.assertEqual(myobj.bar, "Hello")

        # assignment
        myobj["foo"] = "24"                   # Converts to int
        self.assertIsInstance(myobj.foo, int)
        self.assertEqual(myobj.foo, 24)

        myobj.foo = "25"                      # Converts to int
        self.assertIsInstance(myobj["foo"], int)
        self.assertEqual(myobj["foo"], 25)

        with self.assertRaises(ValueError):
            myobj["foo"] = "Not an int"       # Raises exception
        with self.assertRaises(ValueError):
            myobj.foo = "Not an int"          # Raises exception

        # set and update methods
        myobj.set("foo", "26")
        self.assertIsInstance(myobj.foo, int)
        self.assertEqual(myobj.foo, 26)

        myobj.update(foo="27")
        self.assertIsInstance(myobj.foo, int)
        self.assertEqual(myobj.foo, 27)

        myobj.foo = 28
        myobj.setdefault("foo", 29)
        self.assertIsInstance(myobj.foo, int)
        self.assertEqual(myobj.foo, 28)
        del myobj["foo"]
        myobj.setdefault("foo", "30")
        self.assertIsInstance(myobj.foo, int)
        self.assertEqual(myobj.foo, 30)

        with self.assertRaises(ValueError):
            myobj.set("foo", "Not an int")          # Raises exception
        with self.assertRaises(ValueError):
            myobj.update(foo="Not an int")          # Raises exception

        myobj.foo = 24
        myobj.setdefault("foo", "Not an int")       # Not an error (foo already set)
        del myobj["foo"]
        with self.assertRaises(ValueError):
            myobj.setdefault("foo", "Not an int")   # Raises exception if foo unset

        # loading fresh from dict or json
        myobj.fromJSON('{"foo": "32", "bar": "Hello1 "}')
        self.assertEqual(myobj.foo, 32)
        self.assertEqual(myobj.bar, "Hello1")
        myobj = MyObject.newFromJSON('{"foo": "33", "bar": "Hello2 "}')
        self.assertEqual(myobj.foo, 33)
        self.assertEqual(myobj.bar, "Hello2")
        myobj.load_data({"foo": "34", "bar": "Hello3 "})
        self.assertEqual(myobj.foo, 34)
        self.assertEqual(myobj.bar, "Hello3")

        # Not Validated
        myobj.direct_set("foo", "Not an int")     # DANGER: Not an exception!
        self.assertEqual(myobj.foo, "Not an int")
        myobj.foo = 42
        self.assertEqual(myobj.foo, 42)
        myobj.direct_update(foo="Not an int")     # DANGER: Not an exception!
        self.assertEqual(myobj.foo, "Not an int")


    def test_README_serialization(self):
        class ObjSer(Object):
            foo = Attr(int)
            bar = Attr(Obj)

        myobj = Obj(foo="23")
        self.assertEqual(myobj.foo + 1, 24)

        # JSON can be produced and loaded
        myobj = Obj(foo=42, bar="Hello")
        json_string = myobj.toJSON()
        myobj1 = Obj.newFromJSON(json_string)
        self.assertEqual(myobj.dict, myobj1.dict)

        # Even for nested objects
        myobj1 = Obj(foo=12, bar="Hello")
        myobj2 = ObjSer(foo=13, bar=myobj1)
        json_string = myobj2.toJSON()
        myobjn = ObjSer.newFromJSON(json_string)
        self.assertEqual(myobjn.foo, 13)
        self.assertEqual(myobjn.bar.foo, 12)
        self.assertEqual(myobjn.bar.bar, "Hello")

        # Other serialization libraries
        myobj = Obj(foo=42, bar="Hello")
        deflated = marshal.dumps(myobj.deflate_data())
        myobj2 = Obj.inflate_new(marshal.loads(deflated))
        self.assertEqual(myobj.dict, myobj2.dict)
        self.assertTrue(myobj.dict is not myobj2.dict)

        myobj1 = Obj(foo=15, bar="Hello1")
        myobj2 = ObjSer(foo=16, bar=myobj1)
        deflated = marshal.dumps(myobj2.deflate_data())
        myobjn = ObjSer.inflate_new(marshal.loads(deflated))
        self.assertEqual(myobjn.foo, 16)
        self.assertEqual(myobjn.bar.foo, 15)
        self.assertEqual(myobjn.bar.bar, "Hello1")

        # Type hints
        myobj = Obj(foo=23)
        self.assertEqual(
            six.text_type(myobj.toJSON(sort_keys=True)),
            '{{"__class__": "__{}.Obj__", "foo": 23}}'.format(self.__module__)
        )

        # required by default
        myobj = Obj.newFromJSON('{{"__class__": "__{}.Obj__", "foo":23}}'.format(self.__module__))
        self.assertEqual(myobj.foo, 23)

        with six.assertRaisesRegex(self, ValueError, r'should be __{}\.Obj__'.format(self.__module__)):
            Obj.newFromJSON('{"foo":23, "bar":"plugh"}')
        with six.assertRaisesRegex(self, ValueError, r'expected __{}\.Obj__'.format(self.__module__)):
            Obj.newFromJSON('{{"__class__": "__{}.ObjSer__", "foo":23}}'.format(self.__module__))

        # local override
        myobj = Obj.newFromJSON('{"foo":23, "bar":"plugh"}', verifyclass=False)
        self.assertEqual(myobj.foo, 23)
        self.assertEqual(myobj.bar, "plugh")

        class ObjSer1(Object):
            amethyst_includeclass  = False
            amethyst_verifyclass   = False
            foo = Attr(int)
            bar = Attr(isa=six.text_type).strip()

        myobj = ObjSer1.newFromJSON('{"foo":"23", "bar":"plugh"}')
        self.assertEqual(myobj.toJSON(sort_keys=True), '{"bar": "plugh", "foo": 23}')


    def test_immutability(self):
        obj = Obj()

        obj.amethyst_make_immutable()
        self.assertFalse(obj.amethyst_is_mutable(), "is not mutable")

        with self.assertRaises(ImmutableObjectException, msg="Can't set fields when immutable"):
            obj["foo"] = 23

        self.assertIs(obj.amethyst_make_mutable(), obj, "make_mutable returns self")
        self.assertTrue(obj.amethyst_is_mutable(), "is mutable")
        obj["bip"] = 23

        self.assertIs(obj.amethyst_make_immutable(), obj, "make_mutable returns self")
        self.assertFalse(obj.amethyst_is_mutable(), "is not mutable")

        self.assertEqual(obj["bip"], 23, "Can read values when immutable")


    def test_subclass(self):
        obj = Obj(foo=23)
        obj["bar"] = "12"

        self.assertEqual(obj.foo, 23, "Getattr works in subclass when set by constructor")
        self.assertEqual(obj.bar, "12", "Getattr works in subclass when set by setitem")
        self.assertIsNone(obj.baz, "Getattr works on uninitialized values")

        self.assertEqual(obj.dict, {"foo": 23, "bar": "12"}, "No autovivification")

        obj.amethyst_make_immutable()
        self.assertIsNone(obj.bip, "Can read non-existant keys when immutable")

        obj = Object()
        with self.assertRaises(AttributeError, msg="Subclasses don't change parent attributes"):
            obj.foo

        with self.assertRaises(DuplicateAttributeException, msg="Duplicate attribute raises exception"):
            class ObjSub1(Obj):
                foo = Attr(int)

        class ObjSub(Obj):
            jsonhooks = { "__bob__": (lambda obj: "BOB") }
            bab = Attr(int)
            flags = Attr(isa=set)

        self.assertTrue(hasattr(ObjSub, "foo"), "Attrs are inherited")

        obj = ObjSub()
        obj.fromJSON('{{"__class__": "__{}.ObjSub__", "bar": {{"__bob__": "chaz"}}, "flags": {{"__set__": ["chaz"]}}, "baz": "123.45"}}'.format(self.__module__))
        self.assertEqual(obj.bar, "BOB", "jsonhooks extensions")
        self.assertEqual(list(obj.flags)[0], "chaz", "jsonhooks extensions inherit originals")

        obj = Object()
        obj.fromJSON('{"__class__": "__amethyst.core.obj.Object__", "bab": {"__bob__": "chaz"}, "flags": {"__set__": ["chaz"]}, "baz": "123.45"}', import_strategy="sloppy")
        self.assertEqual(obj.get("bab"), {"__bob__": "chaz"}, "jsonhooks extensions do not modify base classes")


    def test_default(self):
        class ObjDflt(Object):
            foo = Attr(int, default=3)
            bar = Attr(default=list)
            baz = Attr(default=[])

        a = ObjDflt()
        b = ObjDflt()

        self.assertEqual(a.foo, 3, "default int a")
        self.assertEqual(b.foo, 3, "default int b")

        self.assertIsInstance(a.bar, list, "default list constructor")
        self.assertIsInstance(a.baz, list, "default list")

        self.assertTrue(a.bar is not b.bar, "default list constructor initializes different objects")
        self.assertTrue(a.baz is b.baz, "default list initializes identical object")


    def test_nested(self):
        class ObjNest(Object):
            foo = Attr(int)
            bar = Attr(Obj)
            baz = Attr(isa=Obj)

        myobj1 = Obj(foo=12, bar="Hello")
        myobj2 = ObjNest(foo=13, bar=myobj1, baz=myobj1)
        self.assertTrue(myobj2.bar is myobj1)            # identical objects
        self.assertTrue(myobj2.bar.dict is myobj1.dict)  # shallowly
        self.assertTrue(myobj2.baz is myobj1)            # isa does not modify


    def test_in_de_flation(self):
        obj = Obj(foo=23, bip=[Obj(baz=2.3), 11])

        dname = "__{}.{}__".format(Obj.__module__, Obj.__name__)
        self.assertIn(dname, ('__test_obj.Obj__', '____main__.Obj__'))

        plain = amethyst_deflate(obj)
        self.assertIsInstance(plain, dict)
        self.assertIsInstance(plain[dname], dict)
        self.assertIsInstance(plain[dname]['foo'], int)
        self.assertIsInstance(plain[dname]['bip'], list)
        self.assertIsInstance(plain[dname]['bip'][0], dict)

        new = amethyst_inflate(plain)
        self.assertIsInstance(new, Obj)
        self.assertEqual(new.foo, 23)
        self.assertIsInstance(new.bip[0], Obj)
        self.assertEqual(new.bip[0].baz, 2.3)
        self.assertEqual(new.bip[1], 11)
        self.assertIsNone(new.bar)
        self.assertIsNone(new.baz)
        self.assertIsNone(new.bip[0].foo)

    def test_integration(self):
        """
        Tests to ensure that we play well with other common libraries.

        six:
          - ensure that we can be passed to six.iteritems()

        sqlite3:
          - ensure that sqlite3.Row objects can be used to initialize objects
        """
        # six.iteritems(obj)
        try:
            import six
            obj = Obj(foo=12, bar="hi", baz=2.3)
            got = set()
            for k, v in six.iteritems(obj):
                got.add("{}={}".format(k, v))
            self.assertEqual(got, set(["foo=12", "bar=hi", "baz=2.3"]), "works with six.iteritems")

        except ImportError:
            raise unittest.SkipTest("six not installed, skipping six integration tests")

        # Object(sqlite3.Row)
        conn = sqlite3.connect(":memory:")
        conn.row_factory = sqlite3.Row
        conn.execute("CREATE TABLE test (foo INTEGER, bar TEXT, baz TEXT, bip TEXT)")
        conn.execute("INSERT INTO test VALUES (74, 'plugh', '12.5', NULL)")

        obj = Obj(conn.execute('SELECT * FROM test').fetchone())
        self.assertEqual(obj.foo, 74)
        self.assertEqual(obj.bar, "plugh")
        self.assertEqual(obj.baz, 12.5)
        self.assertEqual(obj.bip, None)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: LGPL-3.0

from __future__ import division, absolute_import, print_function, unicode_literals
import unittest

import six

import os.path
import sys
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)) + '/../')

from amethyst.core import Object, Attr

class Exception_Eager(Exception): pass
class Counter(object):
    def __init__(self):
        self.value = 0
    def step(self):
        self.value += 1
        return self.value

# package-level object for testing string attrs
class AttrObject(Object):
    foo = Attr("AttrObject")
    bar = Attr(".Counter")

class MyTest(unittest.TestCase):

    def test_basic(self):
        class MyObject(Object):
            amethyst_register_type = False
            foo = Attr(int)
            bar = Attr(isa=six.text_type).strip()

        myobj = MyObject(dict(foo=23, bar=" Hello "))

        self.assertEqual(myobj.foo, 23)
        self.assertEqual(myobj.bar, "Hello")

        myobj.set("foo", 15, foo=5)
        self.assertEqual(myobj.foo, 15, ".set() takes positional or kwargs and prefers positional")


    def test_Attr_standalone(self):
        chk = Attr(int)

        self.assertTrue(callable(chk))
        self.assertEqual(chk("42"), 42)

        with self.assertRaises(TypeError):
            chk(None)
        with self.assertRaises(ValueError):
            chk("plugh")

        for chk in (Attr(int, default=0), Attr(int, default=lambda: 0)):
            self.assertTrue(callable(chk))
            self.assertEqual(chk("42"), 42)

            with self.assertRaises(TypeError):
                chk(None)
            with self.assertRaises(ValueError):
                chk("plugh")
            self.assertEqual(chk.get_default(), 0)


    def test_Attr_classes(self):
        counter = Counter()
        obj1 = AttrObject()
        obj2 = AttrObject(foo=dict(foo=obj1), bar=counter)

        self.assertTrue(isinstance(obj2.foo, AttrObject))
        self.assertTrue(obj2.foo.foo is obj1)
        self.assertTrue(obj2.foo is not obj1)
        self.assertTrue(obj2.bar is counter)


    def test_Attr_unit(self):
        counter = Counter()

        def boom_eager():
            raise Exception_Eager("boom!")

        class MyObject(Object):
            amethyst_verifyclass = False
            amethyst_register_type = False
            a = Attr(int)
            b = Attr(float)
            c = Attr(verify=(lambda x: 0 < x < 10))
            d = Attr(isa=int)
            e = Attr(isa=(int, list))
            f = Attr(default=42)
            g = Attr(default=(lambda: 22))
            h = Attr(default=(lambda: counter.step()))
            i = Attr(builder=(lambda: self.assertFalse(True, "builder called lazily")))
            j = Attr(builder=(lambda: 12))
            k = Attr(float) & Attr(verify=(lambda x: 0 < x < 10))
            l = Attr(float) | Attr(int)  # noqa E741
            m = Attr(int) | Attr(float)
            n = Attr(int) == (2,3,6,7,8)
            o = Attr(int) != (0, 1)

        class MyEagerTest(Object):
            amethyst_verifyclass = False
            amethyst_register_type = False
            k = Attr(default=(lambda: boom_eager()))

        myobj = MyObject()

        self.assertIsNone(myobj.a, "unset is none")

        myobj.set("a", "23")
        self.assertEqual(myobj.a, 23, "Attr(int)")
        with self.assertRaises(ValueError):
            myobj.set("a", "34.2")
        with self.assertRaises(ValueError):
            myobj.set("a", "Not an int")

        myobj.set("b", "23.3")
        self.assertEqual(myobj.b, 23.3, "Attr(float)")

        myobj.set("c", 3.3)
        self.assertEqual(myobj.c, 3.3, "verify 0 < x < 10")
        with self.assertRaises(ValueError):
            myobj.set("c", 12)

        myobj.set("d", 3)
        self.assertEqual(myobj.d, 3, "isa int")
        with self.assertRaises(ValueError):
            myobj.set("c", 12)

        myobj.set(e=23)
        self.assertEqual(myobj.e, 23, "isa tuple:int")
        myobj.set(e=["foo", "bar"])
        self.assertEqual(myobj.e, ["foo", "bar"], "isa tuple:list")
        with self.assertRaises(ValueError):
            myobj.set("e", "Not an int or list")

        self.assertEqual(myobj.f, 42, "default")
        self.assertEqual(myobj.g, 22, "default callable")

        self.assertEqual(myobj.h, 1, "default callable caches (first)")
        self.assertEqual(myobj.h, 1, "default callable caches (second)")

        myobj2 = MyObject()
        myobj3 = MyObject()
        self.assertEqual(myobj2.h, 2, "default callable distinct each object (2)")
        self.assertEqual(myobj3.h, 3, "default callable distinct each object (3)")

        # DO NOT get .i, need to test builder is lazy
        # self.assertEqual(myobj.i, 12, "builder")
        self.assertEqual(myobj.j, 12, "builder")

        myobj.k = "5"
        self.assertEqual(myobj.k, 5)
        with self.assertRaises(ValueError):
            myobj.k = "Not an int or list"
        with self.assertRaises(ValueError):
            myobj.k = 12

        myobj.l = "5"  # noqa E741
        self.assertEqual(myobj.l, 5)
        self.assertTrue(not isinstance(myobj.l, int))
        myobj.l = "5.0"  # noqa E741
        self.assertEqual(myobj.l, 5)
        self.assertTrue(not isinstance(myobj.l, int))
        myobj.m = "5"
        self.assertEqual(myobj.m, 5)
        self.assertTrue(isinstance(myobj.m, int))
        myobj.m = "5.0"
        self.assertEqual(myobj.m, 5)
        self.assertTrue(not isinstance(myobj.m, int))

        myobj.n = "2"
        self.assertEqual(myobj.n, 2)
        myobj.n = 7
        self.assertEqual(myobj.n, 7)
        with self.assertRaises(ValueError):
            myobj.n = 4
        with self.assertRaises(ValueError):
            myobj.n = 12

        myobj.o = "2"
        self.assertEqual(myobj.o, 2)
        myobj.o = 234299993
        self.assertEqual(myobj.o, 234299993)
        with self.assertRaises(ValueError):
            myobj.o = 0
        with self.assertRaises(ValueError):
            myobj.o = 1

        with self.assertRaises(Exception_Eager):
            myobj = MyEagerTest()


    def test_Attr_docs(self):
        # The Attr() docs specify a few specific odd cases:
        class MyObject(Object):
            amethyst_register_type = False
            foo_int      = Attr(int)
            foo_floatint = Attr(float).int()
            foo_posint   = 0 < Attr(int)
            foo_proofint = (0 <= Attr(int)) <= 200     # Parens are necessary!

        myobj = MyObject()

        myobj.set("foo_int", "23")
        self.assertEqual(myobj.foo_int, 23)

        myobj.set("foo_int", 23.6)
        self.assertEqual(myobj.foo_int, 23)

        myobj.set("foo_floatint", "23.3")
        self.assertEqual(myobj.foo_floatint, 23)

        myobj.set("foo_floatint", 23.6)
        self.assertEqual(myobj.foo_floatint, 23)

        myobj.set("foo_posint", "500")
        self.assertEqual(myobj.foo_posint, 500)

        myobj.set("foo_proofint", "150")
        self.assertEqual(myobj.foo_proofint, 150)

        myobj.set("foo_proofint", 200)
        self.assertEqual(myobj.foo_proofint, 200)

        myobj.set("foo_proofint", 0)
        self.assertEqual(myobj.foo_proofint, 0)

        with self.assertRaises(ValueError):
            myobj.set("foo_int", "23.4")
        with self.assertRaises(ValueError):
            myobj.set("foo_floatint", "sdf")
        with self.assertRaises(ValueError):
            myobj.set("foo_posint", 0)
        with self.assertRaises(ValueError):
            myobj.set("foo_posint", -5)
        with self.assertRaises(ValueError):
            myobj.set("foo_proofint", -5)
        with self.assertRaises(ValueError):
            myobj.set("foo_proofint", 201)



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: LGPL-3.0

from __future__ import division, absolute_import, print_function, unicode_literals
import six.moves
import threading
import unittest

import amethyst.core
from amethyst.core import Attr, cached_property, coalesce, identity, get_class
from amethyst.core import set_of, list_of, dict_of

class Foo(object):
    def __init__(self, bar=None):
        self.computed = 0
        self.errors = []
        self.local = threading.local()
        if bar is not None:
            self.bar = bar

    @cached_property
    def bar(self):
        self.computed += 1
        return 42

    @cached_property(delegate="local")
    def baz(self):
        self.computed += 1
        return six.moves._thread.get_ident()

class MyObject(amethyst.core.Object):
    amethyst_register_type = False
    foo = Attr(set_of(int))
    bar = Attr(list_of(float))
    baz = Attr(dict_of('MyObject'))


class MyTest(unittest.TestCase):

    def test_identity(self):
        foo = Foo()
        self.assertTrue(foo is identity(foo))

    def test_get_class(self):
        self.assertTrue(Foo is get_class("Foo"))
        self.assertTrue(Foo is get_class(".Foo"))
        self.assertTrue(Attr is get_class("amethyst.core.Attr"))
        self.assertTrue(Attr is get_class("Attr"))  # Note: anything in scope is accessible

        self.assertTrue(amethyst.core.Object is get_class("amethyst.core.Object"))
        with self.assertRaises(AttributeError):
            get_class("Object")

    def test_attr_structures(self):
        obj = MyObject()
        obj.foo = (1, "23", 23)
        obj.bar = (1, "23", 2.3, 23)
        obj.baz = dict(a=dict(foo=12), b=obj, c=dict(bar=("2.3", 55)))

        self.assertEqual(obj.foo, set((1,23)))
        self.assertEqual(obj.bar, [1, 23, 2.3, 23])
        self.assertTrue(isinstance(obj.baz, dict))
        self.assertTrue(obj.baz['b'] is obj)
        self.assertTrue(isinstance(obj.baz['a'], MyObject))
        self.assertTrue(isinstance(obj.baz['c'], MyObject))
        self.assertTrue(obj.baz['a'] is not obj.baz['b'])
        self.assertTrue(obj.baz['b'] is not obj.baz['c'])
        self.assertTrue(obj.baz['a'] is not obj.baz['c'])
        self.assertEqual(obj.baz['a'].foo, set((12,)))
        self.assertEqual(obj.baz['c'].bar, [2.3, 55])


    def test_coalesce(self):
        self.assertIsNone(coalesce(), "Nothing!")
        self.assertIsNone(coalesce(None, None, None), "None!")
        self.assertEqual(coalesce(42, None, None), 42, "Match first")
        self.assertEqual(coalesce(None, 42, None, None), 42, "Match middle")
        self.assertEqual(coalesce(None, None, None, "42"), "42", "Match last")
        self.assertEqual(coalesce(None, 0, None, None), 0, "Match zero")
        self.assertFalse(coalesce(None, False, None, None), "Match falsey")

    def _thread_test_cached_property(self, foo, ident, computed):
        # runs in thread, so have to save any exceptions to throw later
        try:
            self.assertEqual(foo.computed, computed)
            self.assertNotEqual(foo.baz, ident)
            self.assertEqual(foo.computed, computed + 1)
            self.assertNotEqual(foo.baz, ident)
            self.assertEqual(foo.computed, computed + 1)
            foo.thread_ok = True
        except Exception as err:
            foo.errors.append(err)

    def test_cached_property(self):
        foo = Foo()
        computed = 0

        self.assertEqual(foo.computed, computed, "Not calculated yet")
        computed += 1
        self.assertEqual(foo.bar, 42, "Computed")
        self.assertEqual(foo.computed, computed, "Calculated once")
        self.assertEqual(foo.bar, 42, "Cached")
        self.assertEqual(foo.computed, computed, "Calculated once (still)")

        foo.bar = 12
        self.assertEqual(foo.bar, 12, "Assigned")
        self.assertEqual(foo.computed, computed, "Calculated once (still)")

        del foo.bar
        computed += 1
        self.assertEqual(foo.bar, 42, "Recomputed")
        self.assertEqual(foo.computed, computed, "Calculated twice")

        foo = Foo(bar=12)
        computed = 0
        self.assertEqual(foo.bar, 12, "Assigned")
        self.assertEqual(foo.computed, computed, "Never calculated")

        del foo.bar
        computed += 1
        self.assertEqual(foo.bar, 42, "Cleared / Computed")
        self.assertEqual(foo.computed, computed, "Calculated finally")

        # threading
        computed += 1
        ident = foo.baz
        self.assertTrue(bool(ident))
        self.assertEqual(foo.computed, computed)
        self.assertEqual(foo.baz, ident)
        self.assertEqual(foo.computed, computed)
        self.assertEqual(foo.local.baz, ident, msg="Written to correct attribute")

        computed += 1
        del foo.baz
        self.assertEqual(foo.baz, ident)
        self.assertEqual(foo.computed, computed)

        thr = threading.Thread(target=self._thread_test_cached_property, args=(foo, ident, computed))
        thr.start()
        thr.join()
        for err in foo.errors:
            raise err
        self.assertTrue(foo.thread_ok) # Extra-paranoid check
        computed += 1                  # incremented in thread

        # Unchanged in this thread
        self.assertEqual(foo.baz, ident)
        self.assertEqual(foo.computed, computed)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import os
import re
import sys
sys.path.insert(0, os.path.abspath('..'))

# -- General configuration ------------------------------------------------

project   = u'Amethyst-core'
copyright = u'2018, Dean Serenevy'
author    = u'Dean Serenevy'

extensions     = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary'
]

autodoc_default_flags = [ 'members' ]
autodoc_member_order = 'bysource'
autosummary_generate = True


templates_path = ['_templates']
source_suffix  = '.rst'
master_doc     = 'index'

version = re.search(r'(?m)^__version__\s*=\s*"([\d.]+(?:[\-\+~.]\w+)*)"', open('../amethyst/core/__init__.py').read()).group(1)
release = version

exclude_patterns = [ '_doc', 'Thumbs.db', '.DS_Store' ]


language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#
# today = ''
#
# Else, today_fmt is used as the format for a strftime call.
#
# today_fmt = '%B %d, %Y'

# The reST default role (used for this markup: `text`) to use for all
# documents.
#
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'nature'
# classic'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = dict( )

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
#
# html_title = u'Amethyst-core v0.6.1'

# A shorter title for the navigation bar.  Default is the same as html_title.
#
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#
# html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = [ ]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#
# html_extra_path = []

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
#
# html_last_updated_fmt = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#
# html_additional_pages = {}

# If false, no module index is generated.
#
# html_domain_indices = True

# If false, no index is generated.
#
# html_use_index = True

# If true, the index is split into individual pages for each letter.
#
# html_split_index = False

# If true, links to the reST sources are added to the pages.
#
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr', 'zh'
#
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# 'ja' uses this config value.
# 'zh' user can custom change `jieba` dictionary path.
#
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'Amethyst-coredoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    'papersize': 'letterpaper',
    'pointsize': '11pt',

     # Additional stuff for the LaTeX preamble.
     #
     # 'preamble': '',

     # Latex figure (float) alignment
     #
     # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Amethyst-core.tex', u'Amethyst-core Documentation', author, 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#
# latex_use_parts = False

# If true, show page references after internal links.
#
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
#
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
#
# latex_appendices = []

# It false, will not define \strong, \code, 	itleref, \crossref ... but only
# \sphinxstrong, ..., \sphinxtitleref, ... To help avoid clash with user added
# packages.
#
# latex_keep_old_macro_names = True

# If false, no module index is generated.
#
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'amethyst-core', u'Amethyst-core Documentation', [author], 1)
]

# If true, show URL addresses after external links.
#
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Amethyst-core', u'Amethyst-core Documentation', author, 'Amethyst-core', 'A sober python base library', 'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#
# texinfo_appendices = []

# If false, no module index is generated.
#
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#
# texinfo_no_detailmenu = False



# -*- coding: utf-8 -*-
"""

"""
# SPDX-License-Identifier: LGPL-3.0
from __future__ import division, absolute_import, print_function, unicode_literals
__all__ = """
cached_property
coalesce
dict_of
get_class
identity
list_of
set_of
smartmatch
tupley
""".split()

import importlib
import inspect
import re
import types


def identity(x):
    """The identity function, returns its (single) argument"""
    return x


# Dumb container for python 2 support in list_of() and dict_of()
class _Container(object):
    pass

def list_of(conv, container=list, package=None, frame=1):
    """
    An :py:class:`amethyst.core.obj.Attr` helper function which will
    validate a list of values. Sample usage::

        class MyObject(Object):
            foo = Attr(list_of(float))

        obj = MyObject(foo="23")
        print(obj.foo)            # [ 23 ]        -  a list with an int

        obj.foo = (1, 2, "23")
        print(obj.foo)            # [ 1, 2, 23 ]  -  a list not a tuple


    :param conv: The conversion function or class. If a class, objects in
    the list which are not already objects of this class will be inflated
    using the class. If passed a string, it will be converted to a class
    (or function) using the :py:func:`get_class` function.

    :param container: Constructor which can take a generator and return
    your desired list-like object. For instance, the :py:func:`set_of`
    function passes `container=set`.

    :param package: String or package object to use as the base for
    relative imports. When specified, is passed unmodified to
    :py:func:`get_class`.

    :param frame: Frame depth, as described in :py:func:`get_class`.

    """
    c = _Container()  # Closure variable for python 2 support (don't have "nonlocal")
    if frame and package is None and not callable(conv) and (conv.startswith('.') or '.' not in conv):
        package = inspect.getmodule(inspect.stack()[frame][0])

    def wrapper(thingun):
        # May not pre-compute these to allow list_of("Foo") to be called
        # within the declaration of the Foo class.
        if not getattr(c, 'initialized', False):
            c.initialized = True
            c.conv = conv if callable(conv) else get_class(conv, package=package, frame=None)
            c.conv_is_type = isinstance(c.conv, type)
        return container(
            (x if c.conv_is_type and isinstance(x, c.conv) else c.conv(x))
            for x in tupley(thingun)
        )
    return wrapper


def set_of(conv, package=None, frame=1):
    """
    Like :py:func:`list_of`, but uses a set container rather than a list
    container.
    """
    return list_of(conv, container=set, frame=frame+1)


def dict_of(conv, key_conv=identity, set_key=None, package=None, frame=1):
    """
    An :py:class:`amethyst.core.obj.Attr` helper function which will
    validate a dict of values and optionally keys. Sample usage::

        class MyObject(Object):
            name = Attr()
            foo = Attr(dict_of("MyObject"))

        obj1 = MyObject(name="Alice")
        obj2 = MyObject(foo={ "a": obj1, "b": dict(name="Bob") })

    In the example, ``obj2.foo`` will be a dictionary with two items. Both
    values will be MyObject objects, the "b" item having been auto-inflated.

    .. WARNING::
       The produced attribute value is a normal python dict. Automatic
       inflation only occurs when initially setting the attribute. Normal
       accesses to the attribute dictionary will not validate or
       auto-inflate. For instance, ``obj2.foo["c"] = dict(name="Carol")``
       will store a python dict to key "c", not a MyObject.

    :param conv: The conversion function or class. If a class, values in
    the dict which are not already objects of this class will be inflated
    using the class. If passed a string, it will be converted to a class
    (or function) using the :py:func:`get_class` function.

    :param key_conv: Conversion function for keys, defaults to identity
    function.

    :param set_key: Optional callable passed key name and inflated value
    object. Can be used to set an attribute on the value objects based on
    keys. For instance, we might use ``set_key=lambda k, v: setattr(v,
    "name", v.name or k)`` to set default "name" attributes.

    :param package: String or package object to use as the base for
    relative imports. When specified, is passed unmodified to
    :py:func:`get_class`.

    :param frame: Frame depth, as described in :py:func:`get_class`.

    """
    c = _Container()  # Closure variable for python 2 support (don't have "nonlocal")
    if package is None and (
            (not callable(conv) and (conv.startswith('.') or '.' not in conv))
         or (not callable(key_conv) and (key_conv.startswith('.') or '.' not in key_conv))
    ):
        package = inspect.getmodule(inspect.stack()[frame][0])

    def wrapper(d):
        # May not pre-compute these to allow list_of("Foo") to be called
        # within the declaration of the Foo class.
        if not getattr(c, 'initialized', False):
            c.initialized = True
            c.conv = conv if callable(conv) else get_class(conv, package=package, frame=None)
            c.conv_is_type = isinstance(c.conv, type)
            c.key_conv = key_conv if callable(key_conv) else get_class(key_conv, package=package, frame=None)
            c.key_conv_is_type = isinstance(c.key_conv, type)

        rv = dict()
        for k, v in d.items():
            key = k if c.key_conv_is_type and isinstance(k, c.key_conv) else c.key_conv(k)
            val = v if c.conv_is_type and isinstance(v, c.conv) else c.conv(v)
            if set_key:
                set_key(key, val)
            rv[key] = val
        return rv

    return wrapper


def get_class(name, package=None, frame=1):
    """
    Load a class (or function or other package attribute) from a string
    name. Automatically imports required package. Requested name may be
    relative. If relative and no package is passed, the call stack is
    examined at the frame counter and imports are relative to that package.

    ::

       get_class("foo.Bar")                     # Bar class from package foo
       get_class(".Foo")  or  get_class("Foo")  # Foo class from current package
       get_class(".Foo", frame=2)               # Foo class from caller's package

    :param str name: A string containing a package name and class or object
    name joined by a dot. The package will be loaded and the attribute will
    be returned. From the name of the function, the intention if for
    automatic loading of classes. For example,
    `get_class("amethyst.core.Object")`, however, python doesn't really
    distinguish between loading a class and loading a function or other
    package variable, so the object after the last dot can really be
    anything available in the package -- even unrelated imports to the
    package! For this reason, if classes or packages are imported for user
    or configuration input, it is a good idea to verify that the imported
    object matches some expected base class.

    :param package: String or package object to use as the base for
    relative imports.

    :param frame: Frame depth, for the default base package. When set to 1,
    relative class names are looked up relative to the caller's package.
    When set to a larger value, will look up relative to the caller's
    caller's ... package. Set to 0 or None (or set an explicit value for
    *package* to disable automatically selecting a base package.

    """
    # Rewrite "Foo" as ".Foo"
    if '.' not in name:
        name = '.' + name

    # Split out module and object name
    mod = name[0:name.rindex(".")]
    cls = name[len(mod)+1:]
    pkg = package if isinstance(package, types.ModuleType) else None
    # if name is ".Foo" or "..Foo" we should not have stripped off a "."
    if mod in ('', '.'):
        mod = mod + '.'

    # Inspect stack and get calling package if relative name
    if frame and package is None and name.startswith('.') or '.' not in name:
        try:
            pkg = inspect.getmodule(inspect.stack()[frame][0])
        except Exception:
            pass

    # Special case for ".Foo", we got the package directly
    if mod == '.' and pkg:
        pass
    else:
        if pkg and (package is None or isinstance(package, types.ModuleType)):
            package = pkg.__name__
        pkg = importlib.import_module(mod, package)
    return getattr(pkg, cls)


def tupley(thingun):
    """
    Make sure thingun is like a tuple - a list, set, tuple. If not, wraps
    thingun into a single-item or empty (when None) tuple.
    """
    if thingun is None:
        return ()
    if isinstance(thingun, (list, tuple, set, frozenset)):
        return thingun
    return (thingun,)


def coalesce(*args):
    """
    Returns first argument which is not `None`. If no non-None argumenst,
    then will return `None`. Also returns `None` if argument list is empty.
    """
    for x in args:
        if x is not None:
            return x
    return None


RE_TYPE = type(re.compile("^$"))
NONE_TYPE = type(None)
def smartmatch(val, other):
    """
    Smart match against a value

    Convenient function to use in attribute validation. Attempts to
    determine if a value is like other values. Behavior depends on type of
    the other object:

    * `list`, `tuple`, `set`, `frozenset`: Test membership and return the value
      unmodified.

    * `dict`: Look up the item and return the hashed value.

    * compiled `regex`: call ``other.search(val)``. Remember to anchor your
      search if that is desired!

    * `callable`: call ``other(val)`` and return the result

    * `type`, `NoneType`: Test ``val is other`` and, if true, return value

    * anything else: Test ``val == other`` and, if true, return value

    If none of the above match, raises a :py:exc:`ValueError`
    """
    if isinstance(other, (list, tuple, set, frozenset)):
        if val in other:
            return val

    elif isinstance(other, dict):
        if val in other:
            return other[val]

    elif isinstance(other, RE_TYPE):
        if other.search(val):
            return val

    elif callable(other):
        return other(val)

    elif isinstance(other, (type, NONE_TYPE)):
        if val is other:
            return val

    elif val == other:
        return val

    raise ValueError("Invalid Value")


class cached_property(object):
    """
    Lazy Attribute Memoization

    Creates properties with deferred calculation. Once calculated, the
    result is stored and returned from cache on subsequent access. Useful
    for expensive operations which may not be needed, or to ensure
    just-in-time construction (I like using this for database connections
    or building subwidgets in GUI classes, see examples below).

    Decorator Usage (most common)::

        class Foo(object):
            @cached_property
            def bar(self):
                print("Computing...")
                return 42   # or expensive_calculation()

        foo = Foo()

        print(foo.bar)      # Computing...  42
        print(foo.bar)      # 42

        foo.bar = 12
        print(foo.bar)      # 12

        del foo.bar         # Clears the cache
        print(foo.bar)      # Computing...  42

    Direct use allows calculation to be closure or dynamically chosen.
    The bar attribute will behave the same as above::

        class Foo(object):
            def __init__(self, **kwargs):
                def expensive_calculation():
                    ...

                self.bar = cached_property(expensive_calculation, "bar")


    Example: Automatic, thread-safe, database connections::

        import threading
        import sqlite3
        from amethyst.core import cached_property

        class MyObject(object):
            def __init__(self, **kwargs):
                self._thr_local = threading.local()

            @cached_property(delegate="_thr_local")
            def db(self):
                conn = sqlite3.connect("mydb.sqlite3")
                conn.execute("PRAGMA foreign_keys=ON")
                return conn

        # obj.db will be a different connection in each thread
        # and will only connect if used in the thread


    Example: GUI widget building::

        import wx
        from amethyst.core import cached_property as widget

        class SimpleWindow(wx.Frame):
            def __init__(self, *args, **kwargs):
                super(SimpleWindow, self).__init__(*args, **kwargs)
                self.sizer.Add(self.button1)
                self.sizer.Add(self.button_exit)

            @widget
            def sizer(self):
                widget = wx.BoxSizer(wx.VERTICAL)
                self.SetSizer(widget)
                return widget

            @widget
            def button1(self):
                widget = wx.Button(self, wx.ID_ANY, "Do Something")
                widget.Bind(wx.EVT_BUTTON, self.on_click1)
                return widget

            @widget
            def button_exit(self):
                widget = wx.Button(self, wx.ID_ANY, "Exit")
                widget.Bind(wx.EVT_BUTTON, lambda evt: wx.Exit())
                return widget

            def on_click1(self, evt):
                print("Ouch!")

        class MyApp(wx.App):
            def OnInit(self):
                self.mainwindow.Show(True)
                self.SetTopWindow(self.mainwindow)
                return True

            @widget
            def mainwindow(self):
                return SimpleWindow(None, -1, "This is a test")

        app = MyApp(0)
        app.MainLoop()
    """
    def __init__(self, meth=None, name=None, delegate=None):
        """
        :param meth: The method being decorated. Typically not passed to
           the constructor explicitly, see examples.

        :param name: Key name to use in object dict (or delegate attribute
           name). Automatically extracted from decorated method name if not
           specified.

        :param delegate: Attribute name containing an object to delegate
           storage to. If not `None`, the `name` attribute of `delegate`
           will be accessed (via `getattr`, `setattr`, and `delattr`) when
           determining whether to recompute the cached property and to
           store the computed property value (see example).
        """
        self.name = name
        self.delegate = delegate
        # Simplify implementations by just coding different methods.
        # Python name mangling prevents setting self.__xxx__ directly.
        if delegate:
            self._get = self.get_delegate
            self._set = self.set_delegate
            self._del = self.del_delegate
        else:
            self._get = self.get_obj_dict
            self._set = self.set_obj_dict
            self._del = self.del_obj_dict
        if meth is not None:
            self(meth)

    def __call__(self, meth):
        self.meth = meth
        if self.name is None:
            self.name = meth.__name__
        return self

    def __get__(self, obj, typ=None):
        return self._get(obj, typ)
    def __set__(self, obj, value):
        self._set(obj, value)
    def __delete__(self, obj):
        self._del(obj)

    # object-dict storage
    def get_obj_dict(self, obj, typ=None):
        if self.name not in obj.__dict__:
            obj.__dict__[self.name] = self.meth(obj)
        return obj.__dict__[self.name]

    def set_obj_dict(self, obj, value):
        obj.__dict__[self.name] = value

    def del_obj_dict(self, obj):
        if self.name in obj.__dict__:
            del obj.__dict__[self.name]

    # delegate storage
    def get_delegate(self, obj, typ=None):
        delegate = getattr(obj, self.delegate)
        try:
            return getattr(delegate, self.name)
        except AttributeError:
            pass
        rv = self.meth(obj)
        setattr(delegate, self.name, rv)
        return rv

    def set_delegate(self, obj, value):
        setattr(getattr(obj, self.delegate), self.name, value)

    def del_delegate(self, obj):
        delattr(getattr(obj, self.delegate), self.name)



#!/usr/bin/env python
"""
A sober python base library
"""
# SPDX-License-Identifier: LGPL-3.0

import io
import re
import setuptools
import unittest

__version__ = re.search(r'(?m)^__version__\s*=\s*"([\d.]+(?:[\-\+~.]\w+)*)"', open('amethyst/core/__init__.py').read()).group(1)

def my_test_suite():
    return unittest.TestLoader().discover('tests', pattern='test_*.py')

with io.open('README.rst', encoding='UTF-8') as fh:
    readme = fh.read()

setuptools.setup(
    name         = 'amethyst-core',
    version      = __version__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
        ],
    url          = 'https://github.com/duelafn/python-amethyst-core',
    author       = "Dean Serenevy",
    author_email = 'dean@serenevy.net',
    description  = "A sober python base library",
    long_description = readme,
    packages     = setuptools.find_packages(),
    requires     = [ "six", ],
    namespace_packages = [ 'amethyst' ],
    test_suite   = 'setup.my_test_suite',
)
